"use strict";
/**
 * GitDocumentDB plugin for remote connection using NodeGit
 * Copyright (c) Hidekazu Kubota
 *
 * This source code is licensed under the Mozilla Public License Version 2.0
 * found in the LICENSE file in the root directory of this source tree.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.push = exports.fetch = exports.checkFetch = exports.clone = exports.name = exports.type = exports.sleep = void 0;
const fs_1 = __importDefault(require("fs"));
const nodegit_1 = __importDefault(require("nodegit"));
const isomorphic_git_1 = __importDefault(require("isomorphic-git"));
const tslog_1 = require("tslog");
const git_documentdb_remote_errors_1 = require("git-documentdb-remote-errors");
const authentication_1 = require("./authentication");
/**
 * @internal
 */
function sleep(msec) {
    return new Promise((resolve) => setTimeout(resolve, msec));
}
exports.sleep = sleep;
/**
 * @public
 */
// eslint-disable-next-line @typescript-eslint/naming-convention
exports.type = 'remote';
/**
 * @public
 */
// eslint-disable-next-line @typescript-eslint/naming-convention
exports.name = 'nodegit';
/**
 * Clone
 *
 * @throws {@link InvalidURLFormatError}
 * @throws {@link NetworkError}
 * @throws {@link HTTPError401AuthorizationRequired}
 * @throws {@link HTTPError404NotFound}
 * @throws {@link CannotConnectError}
 *
 * @throws {@link HttpProtocolRequiredError} (from createCredentialForGitHub)
 * @throws {@link InvalidRepositoryURLError} (from createCredentialForGitHub)
 * @throws {@link InvalidSSHKeyPathError} (from createCredentialForSSH)
 *
 * @throws {@link InvalidAuthenticationTypeError} (from createCredential)
 */
// eslint-disable-next-line complexity
async function clone(workingDir, remoteOptions, logger) {
    logger !== null && logger !== void 0 ? logger : (logger = new tslog_1.Logger({
        name: 'plugin-nodegit',
        minLevel: 'trace',
        displayDateTime: false,
        displayFunctionName: false,
        displayFilePath: 'hidden',
    }));
    logger.debug(`remote-nodegit: clone: ${remoteOptions.remoteUrl}`);
    const res = await nodegit_1.default.Clone.clone(remoteOptions.remoteUrl, workingDir, {
        fetchOpts: {
            callbacks: authentication_1.createCredentialCallback(remoteOptions),
        },
    }).catch((err) => err);
    let error = '';
    if (res instanceof Error) {
        error = res.message;
    }
    // if (error !== 'undefined') console.warn('connect fetch error: ' + error);
    switch (true) {
        case error === '':
            break;
        case error.startsWith('unsupported URL protocol'):
        case error.startsWith('malformed URL'):
            throw new git_documentdb_remote_errors_1.InvalidURLFormatError(error);
        // NodeGit throws them when network is limited.
        case error.startsWith('failed to send request'):
        case error.startsWith('failed to resolve address'):
            throw new git_documentdb_remote_errors_1.NetworkError(error);
        case error.startsWith('unexpected HTTP status code: 401'): // 401 on Ubuntu
        case error.startsWith('request failed with status code: 401'): // 401 on Windows
        case error.startsWith('Method connect has thrown an error'):
        case error.startsWith('remote credential provider returned an invalid cred type'): // on Ubuntu
        case error.startsWith('Failed to retrieve list of SSH authentication methods'):
        case error.startsWith('too many redirects or authentication replays'):
            throw new git_documentdb_remote_errors_1.HTTPError401AuthorizationRequired(error);
        case error.startsWith('unexpected HTTP status code: 404'): // 404 on Ubuntu
        case error.startsWith('request failed with status code: 404'): // 404 on Windows
            throw new git_documentdb_remote_errors_1.HTTPError404NotFound(error);
        default:
            throw new git_documentdb_remote_errors_1.CannotConnectError(error);
    }
}
exports.clone = clone;
/**
 * Check connection by FETCH
 *
 * @throws {@link InvalidGitRemoteError}
 * @throws {@link InvalidURLFormatError}
 * @throws {@link NetworkError}
 * @throws {@link HTTPError401AuthorizationRequired}
 * @throws {@link HTTPError404NotFound}
 * @throws {@link CannotConnectError}
 *
 * @throws {@link HttpProtocolRequiredError} (from createCredentialForGitHub)
 * @throws {@link InvalidRepositoryURLError} (from createCredentialForGitHub)
 * @throws {@link InvalidSSHKeyPathError} (from createCredentialForSSH)
 *
 * @throws {@link InvalidAuthenticationTypeError} (from createCredential)
 *
 * @public
 */
// eslint-disable-next-line complexity
async function checkFetch(workingDir, options, remoteName = 'origin', logger) {
    logger !== null && logger !== void 0 ? logger : (logger = new tslog_1.Logger({
        name: 'plugin-nodegit',
        minLevel: 'trace',
        displayDateTime: false,
        displayFunctionName: false,
        displayFilePath: 'hidden',
    }));
    const callbacks = authentication_1.createCredentialCallback(options);
    const repos = await nodegit_1.default.Repository.open(workingDir);
    // Get NodeGit.Remote
    const remote = await nodegit_1.default.Remote.lookup(repos, remoteName).catch(() => { });
    if (remote === undefined) {
        throw new git_documentdb_remote_errors_1.InvalidGitRemoteError(`remote '${remoteName}' does not exist`);
    }
    const error = String(await remote
        .connect(0 /* FETCH */, callbacks)
        .catch((err) => err));
    await remote.disconnect();
    // if (error !== 'undefined') console.warn('connect fetch error: ' + error);
    switch (true) {
        case error === 'undefined':
            break;
        case error.startsWith('Error: unsupported URL protocol'):
        case error.startsWith('Error: malformed URL'):
            throw new git_documentdb_remote_errors_1.InvalidURLFormatError(error);
        // NodeGit throws them when network is limited.
        case error.startsWith('Error: failed to send request'):
        case error.startsWith('Error: failed to resolve address'):
            throw new git_documentdb_remote_errors_1.NetworkError(error);
        case error.startsWith('Error: unexpected HTTP status code: 401'): // 401 on Ubuntu
        case error.startsWith('Error: request failed with status code: 401'): // 401 on Windows
        case error.startsWith('Error: Method connect has thrown an error'):
        case error.startsWith('Error: remote credential provider returned an invalid cred type'): // on Ubuntu
        case error.startsWith('Failed to retrieve list of SSH authentication methods'):
        case error.startsWith('Error: too many redirects or authentication replays'):
            throw new git_documentdb_remote_errors_1.HTTPError401AuthorizationRequired(error);
        case error.startsWith('Error: unexpected HTTP status code: 404'): // 404 on Ubuntu
        case error.startsWith('Error: request failed with status code: 404'): // 404 on Windows
            throw new git_documentdb_remote_errors_1.HTTPError404NotFound(error);
        default:
            throw new git_documentdb_remote_errors_1.CannotConnectError(error);
    }
    return true;
}
exports.checkFetch = checkFetch;
/**
 * git fetch
 *
 * @throws {@link InvalidGitRemoteError}
 * @throws {@link InvalidURLFormatError}
 * @throws {@link NetworkError}
 * @throws {@link HTTPError401AuthorizationRequired}
 * @throws {@link HTTPError404NotFound}
 * @throws {@link CannotConnectError}
 *
 * @throws {@link HttpProtocolRequiredError} (from createCredentialForGitHub)
 * @throws {@link InvalidRepositoryURLError} (from createCredentialForGitHub)
 * @throws {@link InvalidSSHKeyPathError} (from createCredentialForSSH)
 *
 * @throws {@link InvalidAuthenticationTypeError} (from createCredential)
 *
 * @public
 */
// eslint-disable-next-line complexity
async function fetch(workingDir, remoteOptions, remoteName = 'origin', logger) {
    logger !== null && logger !== void 0 ? logger : (logger = new tslog_1.Logger({
        name: 'plugin-nodegit',
        minLevel: 'trace',
        displayDateTime: false,
        displayFunctionName: false,
        displayFilePath: 'hidden',
    }));
    logger.debug(`remote-nodegit: fetch: ${remoteOptions.remoteUrl}`);
    const repos = await nodegit_1.default.Repository.open(workingDir);
    const callbacks = authentication_1.createCredentialCallback(remoteOptions);
    const res = await repos
        .fetch(remoteName, {
        callbacks,
    })
        .catch((err) => err);
    // It leaks memory if not cleanup
    repos.cleanup();
    let error;
    if (res !== undefined && res !== null) {
        error = res.message;
    }
    // if (error !== 'undefined') console.warn('connect fetch error: ' + error);
    switch (true) {
        case error === undefined || error === null:
            break;
        case /^remote '.+?' does not exist/.test(error):
            throw new git_documentdb_remote_errors_1.InvalidGitRemoteError(error);
        case error.startsWith('unsupported URL protocol'):
        case error.startsWith('malformed URL'):
            throw new git_documentdb_remote_errors_1.InvalidURLFormatError(error);
        // NodeGit throws them when network is limited.
        case error.startsWith('failed to send request'):
        case error.startsWith('failed to resolve address'):
            throw new git_documentdb_remote_errors_1.NetworkError(error);
        case error.startsWith('unexpected HTTP status code: 401'): // 401 on Ubuntu
        case error.startsWith('request failed with status code: 401'): // 401 on Windows
        case error.startsWith('Method connect has thrown an error'):
        case error.startsWith('remote credential provider returned an invalid cred type'): // on Ubuntu
        case error.startsWith('Failed to retrieve list of SSH authentication methods'):
        case error.startsWith('too many redirects or authentication replays'):
            throw new git_documentdb_remote_errors_1.HTTPError401AuthorizationRequired(error);
        case error.startsWith('unexpected HTTP status code: 404'): // 404 on Ubuntu
        case error.startsWith('request failed with status code: 404'): // 404 on Windows
            throw new git_documentdb_remote_errors_1.HTTPError404NotFound(error);
        default:
            throw new git_documentdb_remote_errors_1.CannotConnectError(error);
    }
}
exports.fetch = fetch;
/**
 * Calc distance
 *
 * @internal
 */
function calcDistance(baseCommitOid, localCommitOid, remoteCommitOid) {
    if (baseCommitOid === undefined) {
        return {
            ahead: undefined,
            behind: undefined,
        };
    }
    return {
        ahead: localCommitOid !== baseCommitOid ? 1 : 0,
        behind: remoteCommitOid !== baseCommitOid ? 1 : 0,
    };
}
/**
 * git push
 *
 * @throws {@link InvalidGitRemoteError}
 * @throws {@link UnfetchedCommitExistsError}
 * @throws {@link InvalidURLFormatError}
 * @throws {@link NetworkError}
 * @throws {@link HTTPError401AuthorizationRequired}
 * @throws {@link HTTPError404NotFound}
 * @throws {@link HTTPError403Forbidden}
 * @throws {@link CannotConnectError}
 *
 * @throws {@link UnfetchedCommitExistsError} (from validatePushResult())
 * @throws {@link CannotConnectError} (from validatePushResult())
 *
 * @throws {@link HttpProtocolRequiredError} (from createCredentialForGitHub)
 * @throws {@link InvalidRepositoryURLError} (from createCredentialForGitHub)
 * @throws {@link InvalidSSHKeyPathError} (from createCredentialForSSH)
 *
 * @throws {@link InvalidAuthenticationTypeError} (from createCredential)
 *
 * @public
 */
// eslint-disable-next-line complexity
async function push(workingDir, remoteOptions, remoteName = 'origin', localBranchName = 'main', remoteBranchName = 'main', logger) {
    logger !== null && logger !== void 0 ? logger : (logger = new tslog_1.Logger({
        name: 'plugin-nodegit',
        minLevel: 'trace',
        displayDateTime: false,
        displayFunctionName: false,
        displayFilePath: 'hidden',
    }));
    logger.debug(`remote-nodegit: push: ${remoteOptions.remoteUrl}`);
    const repos = await nodegit_1.default.Repository.open(workingDir);
    const remote = await repos
        .getRemote(remoteName)
        .catch((err) => {
        if (/^remote '.+?' does not exist/.test(err.message)) {
            throw new git_documentdb_remote_errors_1.InvalidGitRemoteError(err.message);
        }
        throw err;
    });
    const callbacks = authentication_1.createCredentialCallback(remoteOptions);
    const localBranch = 'refs/heads/' + localBranchName;
    const remoteBranch = 'refs/heads/' + remoteBranchName;
    const res = await remote
        .push([`${localBranch}:${remoteBranch}`], {
        callbacks,
    })
        .catch((err) => {
        if (err.message.startsWith('cannot push because a reference that you are trying to update on the remote contains commits that are not present locally')) {
            throw new git_documentdb_remote_errors_1.UnfetchedCommitExistsError();
        }
        return err;
    })
        .finally(() => {
        // It leaks memory if not cleanup
        repos.cleanup();
    });
    let error = '';
    if (typeof res !== 'number' && typeof res !== 'undefined') {
        error = res.message;
    }
    // console.warn('connect push error: ' + error);
    switch (true) {
        case typeof res === 'number' || typeof res === 'undefined':
            break;
        case error.startsWith('unsupported URL protocol'):
        case error.startsWith('malformed URL'):
            throw new git_documentdb_remote_errors_1.InvalidURLFormatError(error);
        // NodeGit throws them when network is limited.
        case error.startsWith('failed to send request'):
        case error.startsWith('failed to resolve address'):
            throw new git_documentdb_remote_errors_1.NetworkError(error);
        case error.startsWith('unexpected HTTP status code: 401'): // 401 on Ubuntu
        case error.startsWith('request failed with status code: 401'): // 401 on Windows
        case error.startsWith('Method connect has thrown an error'):
        case error.startsWith('remote credential provider returned an invalid cred type'): // on Ubuntu
        case error.startsWith('Failed to retrieve list of SSH authentication methods'):
        case error.startsWith('too many redirects or authentication replays'):
            throw new git_documentdb_remote_errors_1.HTTPError401AuthorizationRequired(error);
        case error.startsWith('unexpected HTTP status code: 404'): // 404 on Ubuntu
        case error.startsWith('request failed with status code: 404'): // 404 on Windows
            throw new git_documentdb_remote_errors_1.HTTPError404NotFound(error);
        case error.startsWith('unexpected HTTP status code: 403'): // 403 on Ubuntu
        case error.startsWith('request failed with status code: 403'): // 403 on Windows
        case error.startsWith('Error: ERROR: Permission to'): {
            throw new git_documentdb_remote_errors_1.HTTPError403Forbidden(error);
        }
        default:
            throw new git_documentdb_remote_errors_1.CannotConnectError(error);
    }
    await validatePushResult(repos, workingDir, remoteName, remoteBranchName, callbacks);
}
exports.push = push;
/**
 * NodeGit.Remote.push does not throw following error in race condition:
 * 'cannot push because a reference that you are trying to update on the remote contains commits that are not present locally'
 * So check remote changes again.
 *
 * @throws {@link CannotConnectError}
 * @throws {@link UnfetchedCommitExistsError}
 */
async function validatePushResult(repos, workingDir, remoteName, remoteBranchName, callbacks) {
    await repos
        .fetch(remoteName, {
        callbacks,
    })
        .catch((err) => {
        // push() already check errors except network errors.
        // So throw only network errors here.
        throw new git_documentdb_remote_errors_1.CannotConnectError(err.message);
    })
        .finally(() => {
        repos.cleanup();
    });
    // Use isomorphic-git to avoid memory leak
    const localCommitOid = await isomorphic_git_1.default.resolveRef({
        fs: fs_1.default,
        dir: workingDir,
        ref: 'HEAD',
    });
    const remoteCommitOid = await isomorphic_git_1.default.resolveRef({
        fs: fs_1.default,
        dir: workingDir,
        ref: `refs/remotes/${remoteName}/${remoteBranchName}`,
    });
    const [baseCommitOid] = await isomorphic_git_1.default.findMergeBase({
        fs: fs_1.default,
        dir: workingDir,
        oids: [localCommitOid, remoteCommitOid],
    });
    const distance = await calcDistance(baseCommitOid, localCommitOid, remoteCommitOid);
    if (distance.behind && distance.behind > 0) {
        throw new git_documentdb_remote_errors_1.UnfetchedCommitExistsError();
    }
}
//# sourceMappingURL=remote-nodegit.js.map